@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999, 2011 Tanuki Software, Ltd.
+ * Copyright (c) 1999, 2012 Tanuki Software, Ltd.
  * http://www.tanukisoftware.comment
  * All rights reserved.
  *
@@ -57,6 +57,7 @@
 #include "logger.h"
 #include "property.h"
 #include "wrapper.h"
+#include "wrapper_file.h"
 
 #define MAX_INCLUDE_DEPTH 10
 
@@ -642,427 +643,27 @@ int getEncodingByName(char* encodingMB, char** encoding) {
     return FALSE;
 }
 
-/**
- * Loads the contents of a file into the specified properties.
- *  Whenever a line which starts with #include is encountered, then the rest
- *  the line will be interpreted as a cascading include file.  If the file
- *  does not exist, the include definition is ignored.
- *
- * @return PROP_LOAD_SUCCESS if the file was loaded successfully,
- *         PROP_LOAD_FAIL if there were any problems at all, or
- *         PROP_LOAD_HARD_FAIL if the problem should cascaded all the way up.
- */
-#define PROP_LOAD_SUCCESS   101
-#define PROP_LOAD_FAIL      102
-#define PROP_LOAD_HARD_FAIL 103
-int loadPropertiesInner(Properties* properties, const TCHAR* filename, int fileRequired, int depth, const TCHAR* parentFilename, int parentLineNumber) {
-    FILE *stream;
-    char bufferMB[MAX_PROPERTY_NAME_VALUE_LENGTH];
-    TCHAR expBuffer[MAX_PROPERTY_NAME_VALUE_LENGTH];
-    TCHAR *trimmedBuffer;
-    size_t trimmedBufferLen;
-    TCHAR *c;
+static int loadPropertiesCallback(void *callbackParam, const TCHAR *fileName, int lineNumber, TCHAR *config, int debugProperties)
+{
+    Properties *properties = (Properties *)callbackParam;
     TCHAR *d;
-    size_t i, j;
-    size_t len;
-    int quoted;
-    TCHAR *absoluteBuffer;
-    int hadBOM;
-    int lineNumber;
 
-    char *encodingMB;
-#ifdef WIN32
-    int encoding;
-#else
-    char* encoding;
-    char* interumEncoding;
-#endif
-    int includeRequired;
-    int loadResult = PROP_LOAD_SUCCESS;
-    int ret;
-    TCHAR *bufferW;
-#ifdef WIN32
-    int size;
-#endif
+    properties->debugProperties = debugProperties;
 
-#ifdef _DEBUG
-    log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT("loadPropertiesInner(props, '%s', %d, %d, '%s', %d)"),
-        filename, fileRequired, depth, (parentFilename ? parentFilename : TEXT("<NULL>")), parentLineNumber);
-#endif
-
-    /* Look for the specified file. */
-    if ((stream = _tfopen(filename, TEXT("rb"))) == NULL) {
-        /* Unable to open the file. */
-        if (properties->debugIncludes || fileRequired) {
-            if (depth > 0) {
-                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
-                    TEXT("%sIncluded configuration file not found: %s\n  Referenced from: %s (line %d)\n  Current working directory: %s"),
-                    (properties->debugIncludes ? TEXT("  ") : TEXT("")), filename, parentFilename, parentLineNumber, wrapperData->originalWorkingDir);
-            } else {
-                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
-                    TEXT("Configuration file not found: %s\n  Current working directory: %s"), filename, wrapperData->originalWorkingDir);
-            }
-        } else {
-#ifdef _DEBUG
-            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT("Properties file not found: %s"), filename);
-#endif
-        }
-        return PROP_LOAD_FAIL;
-    }
-    if (properties->debugIncludes) {
-        if (!properties->preload) {
-            if (depth > 0) {
-                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
-                    TEXT("  Loading included configuration file, %s"), filename);
-            } else {
-                /* Will not actually get here because the debug includes can't be set until it is loaded.
-                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
-                    TEXT("Loading configuration file, %s"), filename);
-                */
-            }
-        }
+    if (_tcsstr(config, TEXT("include")) == config) {
+        /* Users sometimes remove the '#' from include statements.
+           Add a warning to help them notice the problem. */
+        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ADVICE,
+                   TEXT("Include file reference missing leading '#': %s"), config);
+    } else if ((d = _tcschr(config, TEXT('='))) != NULL) {
+        /* Locate the first '=' in the line, ignore lines that do not contain a '=' */
+        /* Null terminate the first half of the line. */
+        *d = TEXT('\0');
+        d++;
+        addProperty(properties, fileName, lineNumber, config, d, FALSE, FALSE, TRUE, FALSE);
     }
 
-    /* Load in the first row of properties to check the encoding. */
-    if (fgets(bufferMB, MAX_PROPERTY_NAME_VALUE_LENGTH, stream)) {
-        /* If the file starts with a BOM (Byte Order Marker) then we want to skip over it. */
-        if ((bufferMB[0] == (char)0xef) && (bufferMB[1] == (char)0xbb) && (bufferMB[2] == (char)0xbf)) {
-            i = 3;
-            hadBOM = TRUE;
-        } else {
-            i = 0;
-            hadBOM = FALSE;
-        }
-
-        /* Does the file start with "#encoding="? */
-        if ((bufferMB[i++] == '#') && (bufferMB[i++] == 'e') && (bufferMB[i++] == 'n') && (bufferMB[i++] == 'c') &&
-            (bufferMB[i++] == 'o') && (bufferMB[i++] == 'd') && (bufferMB[i++] == 'i') &&
-            (bufferMB[i++] == 'n') && (bufferMB[i++] == 'g') && (bufferMB[i++] == '=')) {
-            encodingMB = bufferMB + i;
-            i = 0;
-            while ((encodingMB[i] != ' ') && (encodingMB[i] != '\n') && (encodingMB[i]  != '\r')) {
-               i++;
-            }
-            encodingMB[i] = '\0';
-
-            if ((hadBOM) && (strIgnoreCaseCmp(encodingMB, "UTF-8") != 0)) {
-            }
-            if (getEncodingByName(encodingMB, &encoding) == TRUE) {
-                fclose(stream);
-                return PROP_LOAD_FAIL;
-            }
-
-        } else {
-#ifdef WIN32
-            encoding = GetACP();
-#else 
-            encoding = nl_langinfo(CODESET);
- #ifdef MACOSX
-            if (strlen(encoding) == 0) {
-                encoding = "UTF-8";
-            }
- #endif
-#endif
-        }
-    } else {
-        /* Failed to read the first line of the file. */
-#ifdef WIN32
-        encoding = GetACP();
-#else 
-        encoding = nl_langinfo(CODESET);
- #ifdef MACOSX
-        if (strlen(encoding) == 0) {
-            encoding = "UTF-8";
-        }
- #endif
-#endif
-    }
-    fclose(stream);
-
-    if ((stream = _tfopen(filename, TEXT("rb"))) == NULL) {
-        /* Unable to open the file. */
-        if (properties->debugIncludes || fileRequired) {
-            if (depth > 0) {
-                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
-                    TEXT("%sIncluded configuration file, %s, was not found."), (properties->debugIncludes ? TEXT("  ") : TEXT("")), filename);
-            } else {
-                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
-                    TEXT("Configuration file, %s, was not found."), filename);
-            }
-
-        } else {
-#ifdef _DEBUG
-            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS, TEXT("Properties file not found: %s"), filename);
-#endif
-        }
-        return PROP_LOAD_FAIL;
-    }
-
-    /* Load in all of the properties */
-    lineNumber = 1;
-    do {
-        c = (TCHAR*)fgets(bufferMB, MAX_PROPERTY_NAME_VALUE_LENGTH, stream);
-        if (c != NULL) {
-#ifdef WIN32
-            ret = multiByteToWideChar(bufferMB, encoding, &bufferW, TRUE);
-#else
-            interumEncoding = nl_langinfo(CODESET);
- #ifdef MACOSX
-            if (strlen(interumEncoding) == 0) {
-                interumEncoding = "UTF-8";
-            }
- #endif
-            ret = multiByteToWideChar(bufferMB, encoding, interumEncoding, &bufferW, TRUE);
-#endif
-            if (ret) {
-                if (bufferW) {
-                    /* bufferW contains an error message. */
-                    if (!properties->preload) {
-                        if (depth > 0) {
-                            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN,
-                                TEXT("%sIncluded configuration file, %s, contains a problem on line #%d and could not be read. (%s)"),
-                                (properties->debugIncludes ? TEXT("  ") : TEXT("")), filename, lineNumber, bufferW);
-                        } else {
-                            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN,
-                                TEXT("Configuration file, %s, contains a problem on line #%d and could not be read. (%s)"), filename, lineNumber, bufferW);
-                        }
-                    }
-                    free(bufferW);
-                } else {
-                    outOfMemory(TEXT("LPI"), 1);
-                }
-                fclose(stream);
-                return PROP_LOAD_FAIL;
-            }
-            
-#ifdef _DEBUG
-            /* The line feeds are not yet stripped here. */
-            /*
- #ifdef WIN32
-            wprintf(TEXT("%s:%d (%d): [%s]\n"), filename, lineNumber, encoding, bufferW);
- #else
-            wprintf(TEXT("%S:%d (%s to %s): [%S]\n"), filename, lineNumber, encoding, interumEncoding, bufferW);
- #endif
-            */
-#endif
-            
-            c = bufferW;
-            /* Always strip both ^M and ^J off the end of the line, this is done rather
-             *  than simply checking for \n so that files will work on all platforms
-             *  even if their line feeds are incorrect. */
-            if ((d = _tcschr(bufferW, 0x0d /* ^M */)) != NULL) {
-                d[0] = TEXT('\0');
-            }
-            if ((d = _tcschr(bufferW, 0x0a /* ^J */)) != NULL) {
-                d[0] = TEXT('\0');
-            }
-            /* Strip any whitespace from the front of the line. */
-            trimmedBuffer = bufferW;
-            while ((trimmedBuffer[0] == TEXT(' ')) || (trimmedBuffer[0] == 0x08)) {
-                trimmedBuffer++;
-            }
-
-            /* If the line does not start with a comment, make sure that
-             *  any comment at the end of line are stripped.  If any any point, a
-             *  double hash, '##', is encountered it should be interpreted as a
-             *  hash in the actual property rather than the beginning of a comment. */
-            if (trimmedBuffer[0] != TEXT('#')) {
-                len = _tcslen(trimmedBuffer);
-                i = 0;
-                quoted = 0;
-                while (i < len) {
-                    if (trimmedBuffer[i] == TEXT('"')) {
-                        quoted = !quoted;
-                    } else if ((trimmedBuffer[i] == TEXT('#')) && (!quoted)) {
-                        /* Checking the next character will always be ok because it will be
-                         *  '\0 at the end of the string. */
-                        if (trimmedBuffer[i + 1] == TEXT('#')) {
-                            /* We found an escaped #. Shift the rest of the string
-                             *  down by one character to remove the second '#'.
-                             *  Include the shifting of the '\0'. */
-                            for (j = i + 1; j <= len; j++) {
-                                trimmedBuffer[j - 1] = trimmedBuffer[j];
-                            }
-                            len--;
-                        } else {
-                            /* We found a comment. So this is the end. */
-                            trimmedBuffer[i] = TEXT('\0');
-                            len = i;
-                        }
-                    }
-                    i++;
-                }
-            }
-
-            /* Strip any whitespace from the end of the line. */
-            trimmedBufferLen = _tcslen(trimmedBuffer);
-            while ((trimmedBufferLen > 0) && ((trimmedBuffer[trimmedBufferLen - 1] == TEXT(' '))
-            || (trimmedBuffer[trimmedBufferLen - 1] == 0x08))) {
-
-                trimmedBuffer[--trimmedBufferLen] = TEXT('\0');
-            }
-
-            /* Only look at lines which contain data and do not start with a '#'
-             *  If the line starts with '#include' then recurse to the include file */
-            if (_tcslen(trimmedBuffer) > 0) {
-                if (strcmpIgnoreCase(trimmedBuffer, TEXT("#include.debug")) == 0) {
-                    /* Enable include file debugging. */
-                    if (properties->preload == FALSE) {
-                        properties->debugIncludes = TRUE;
-                    } else {
-                        properties->debugIncludes = FALSE;
-                    }
-                } else if ((_tcsstr(trimmedBuffer, TEXT("#include ")) == trimmedBuffer) || (_tcsstr(trimmedBuffer, TEXT("#include.required ")) == trimmedBuffer)) {
-                    if (_tcsstr(trimmedBuffer, TEXT("#include.required ")) == trimmedBuffer) {
-                        /* The include file is required. */
-                        includeRequired = TRUE;
-                        c = trimmedBuffer + 18;
-                    } else {
-                        /* Include file, if the file does not exist, then ignore it */
-                        includeRequired = FALSE;
-                        c = trimmedBuffer + 9;
-                    }
-                    
-                    /* Strip any leading whitespace */
-                    while ((c[0] != TEXT('\0')) && (c[0] == TEXT(' '))) {
-                        c++;
-                    }
-
-                    /* The filename may contain environment variables, so expand them. */
-                    if (properties->debugIncludes) {
-                        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
-                            TEXT("Found #include file in %s: %s"), filename, c);
-                    }
-                    evaluateEnvironmentVariables(c, expBuffer, MAX_PROPERTY_NAME_VALUE_LENGTH);
-
-                    if (properties->debugIncludes && (_tcscmp(c, expBuffer) != 0)) {
-                        /* Only show this log if there were any environment variables. */
-                        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
-                            TEXT("  After environment variable replacements: %s"), expBuffer);
-                    }
-
-                    /* Now obtain the real absolute path to the include file. */
-#ifdef WIN32
-                    /* Find out how big the absolute path will be */
-                    size = GetFullPathName(expBuffer, 0, NULL, NULL); /* Size includes '\0' */
-                    if (!size) {
-                        if (properties->debugIncludes || includeRequired) {
-                            log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
-                                TEXT("Unable to resolve the full path of included configuration file: %s (%s)\n  Referenced from: %s (line %d)\n  Current working directory: %s"),
-                                expBuffer, getLastErrorText(), filename, lineNumber, wrapperData->originalWorkingDir);
-                        }
-                        absoluteBuffer = NULL;
-                    } else {
-                        absoluteBuffer = malloc(sizeof(TCHAR) * size);
-                        if (!absoluteBuffer) {
-                            outOfMemory(TEXT("LPI"), 1);
-                        } else {
-                            if (!GetFullPathName(expBuffer, size, absoluteBuffer, NULL)) {
-                                if (properties->debugIncludes || includeRequired) {
-                                    log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
-                                        TEXT("Unable to resolve the full path of included configuration file: %s (%s)\n  Referenced from: %s (line %d)\n  Current working directory: %s"),
-                                        expBuffer, getLastErrorText(), filename, lineNumber, wrapperData->originalWorkingDir);
-                                }
-                                free(absoluteBuffer);
-                                absoluteBuffer = NULL;
-                            }
-                        }
-                    }
-#else
-                    absoluteBuffer = malloc(sizeof(TCHAR) * (PATH_MAX + 1));
-                    if (!absoluteBuffer) {
-                        outOfMemory(TEXT("LPI"), 2);
-                    } else {
-                        if (_trealpath(expBuffer, absoluteBuffer) == NULL) {
-                            if (properties->debugIncludes || includeRequired) {
-                                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
-                                    TEXT("Unable to resolve the full path of included configuration file: %s (%s)\n  Referenced from: %s (line %d)\n  Current working directory: %s"),
-                                    expBuffer, getLastErrorText(), filename, lineNumber, wrapperData->originalWorkingDir);
-                            }
-                            free(absoluteBuffer);
-                            absoluteBuffer = NULL;
-                        }
-                    }
-#endif
-                    if (absoluteBuffer) {
-                        if (depth < MAX_INCLUDE_DEPTH) {
-                            loadResult = loadPropertiesInner(properties, absoluteBuffer, includeRequired, depth + 1, filename, lineNumber);
-                            if (loadResult == PROP_LOAD_SUCCESS) {
-                                /* Ok continue. */
-                            } else if ((loadResult == PROP_LOAD_FAIL) || (loadResult == PROP_LOAD_HARD_FAIL)) {
-                                /* Failed. */
-                                if (includeRequired) {
-                                    /* Include file was required, but we failed to load it. */
-                                    if (!properties->preload) {
-                                        log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN,
-                                            TEXT("%sThe required configuration file, %s, was not loaded.\n%s  Referenced from: %s (line %d)"),
-                                            (properties->debugIncludes ? TEXT("  ") : TEXT("")), absoluteBuffer, (properties->debugIncludes ? TEXT("  ") : TEXT("")), filename, lineNumber);
-                                    }
-                                    loadResult = PROP_LOAD_HARD_FAIL;
-                                }
-                                if (loadResult == PROP_LOAD_HARD_FAIL) {
-                                    /* Can't continue. */
-                                    break;
-                                } else {
-                                    /* Failed but continue. */
-                                    loadResult = PROP_LOAD_SUCCESS;
-                                }
-                            } else {
-                                _tprintf(TEXT("Unexpected load error %d\n"), loadResult);
-                                /* continue. */
-                                loadResult = PROP_LOAD_SUCCESS;
-                            }
-                        } else {
-                            if (properties->debugIncludes) {
-                                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_STATUS,
-                                    TEXT("  Unable to include configuration file, %s, because the max include depth was reached."), absoluteBuffer);
-                            }
-                        }
-                        free(absoluteBuffer);
-                    } else {
-                        if (includeRequired) {
-                            /* Include file was required, but we failed to load it. */
-                            if (!properties->preload) {
-                                log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_WARN,
-                                    TEXT("%sThe required configuration file, %s, was not loaded.\n%s  Referenced from: %s (line %d)"),
-                                    (properties->debugIncludes ? TEXT("  ") : TEXT("")), expBuffer, (properties->debugIncludes ? TEXT("  ") : TEXT("")), filename, lineNumber);
-                            }
-                            loadResult = PROP_LOAD_HARD_FAIL;
-                            break;
-                        }
-                    }
-                } else if (strcmpIgnoreCase(trimmedBuffer, TEXT("#properties.debug")) == 0) {
-                    if (!properties->preload) {
-                        /* Enable property debugging. */
-                        properties->debugProperties = TRUE;
-                    }
-                } else if (_tcsstr(trimmedBuffer, TEXT("include")) == trimmedBuffer) {
-                    /* Users sometimes remove the '#' from include statements.  Add a warning to help them notice the problem. */
-                    log_printf(WRAPPER_SOURCE_WRAPPER, LEVEL_ADVICE,
-                        TEXT("Include file reference missing leading '#': %s"), trimmedBuffer);
-                } else if (trimmedBuffer[0] != TEXT('#')) {
-                    /*_tprintf(TEXT("%s\n"), trimmedBuffer);*/
-
-                    /* Locate the first '=' in the line, ignore lines that do not contain a '=' */
-                    if ((d = _tcschr(trimmedBuffer, TEXT('='))) != NULL) {
-                        /* Null terminate the first half of the line. */
-                        *d = TEXT('\0');
-                        d++;
-                        addProperty(properties, filename, lineNumber, trimmedBuffer, d, FALSE, FALSE, TRUE, FALSE);
-                    }
-                }
-            }
-            
-            /* Always free each line read. */
-            free(bufferW);
-        }
-        lineNumber++;
-    } while (c != NULL);
-
-    /* Close the file */
-    fclose(stream);
-
-    return loadResult;
+    return TRUE;
 }
 
 /**
@@ -1084,28 +685,32 @@ int loadProperties(Properties *properties, const TCHAR* filename, int preload) {
     #endif
     time_t      now;
     struct tm   *nowTM;
+    ConfigFileReader reader;
     int loadResult;
     
-    /* Store the preload flag for this loading of properties. */
-    properties->preload = preload;
-
 #ifdef WIN32
     _ftime( &timebNow );
     now = (time_t)timebNow.time;
-    #else
+#else
     gettimeofday(&timevalNow, NULL);
     now = (time_t)timevalNow.tv_sec;
-    #endif
+#endif
     nowTM = localtime(&now);
     memcpy(&loadPropertiesTM, nowTM, sizeof(struct tm));
 
-    loadResult = loadPropertiesInner(properties, filename, 0, 0, NULL, 0);
+    configFileReader_Initialize(&reader, loadPropertiesCallback, properties, TRUE);
+
+    /* Store the preload flag for this loading of properties. */
+    reader.preload = preload;
+
+    loadResult = configFileReader_Read(&reader, filename, 0, 0, NULL, 0);
+
     /* Any failure is a failure in the root. */
     switch (loadResult) {
-    case PROP_LOAD_SUCCESS:
+    case CONFIG_FILE_READER_SUCCESS:
         return FALSE;
-    case PROP_LOAD_FAIL:
-    case PROP_LOAD_HARD_FAIL:
+    case CONFIG_FILE_READER_FAIL:
+    case CONFIG_FILE_READER_HARD_FAIL:
         return TRUE;
     default:
         _tprintf(TEXT("Unexpected load error %d\n"), loadResult);
@@ -1119,7 +724,6 @@ Properties* createProperties() {
         outOfMemory(TEXT("CP"), 1);
         return NULL;
     }
-    properties->debugIncludes = FALSE;
     properties->debugProperties = FALSE;
     properties->first = NULL;
     properties->last = NULL;
@@ -1866,9 +1470,11 @@ int addPropertyPair(Properties *properties, const TCHAR* filename, int lineNum,
         /* Null terminate the first half of the line. */
         *d = TEXT('\0');
         d++;
-        addProperty(properties, filename, lineNum, buffer, d, finalValue, quotable, FALSE, internal);
-
-        return 0;
+        if (addProperty(properties, filename, lineNum, buffer, d, finalValue, quotable, FALSE, internal) != NULL) {
+            return 0;
+        } else {
+            return 1;
+        }
     } else {
         return 1;
     }
